from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Union, Type

from .botcontext import BotContext
from .botmsg import BotMessage
from .user import User


class InboundMessage(ABC):
    __slots__ = ('_data',)

    _UserClass = User

    def __init__(self, data):
        self._data = data
        self.validate()
        self._user = self._UserClass(self.user_id)

    def __getitem__(self, key):
        return self._data[key]

    @property
    @abstractmethod
    def text(self) -> str:
        pass

    # region User
    @property
    def user(self) -> User:
        return self._user

    @property
    @abstractmethod
    def user_id(self) -> Union[int, str]:
        pass

    @property
    @abstractmethod
    def user_id_url(self) -> str:
        pass

    @property
    def user_visual_name(self) -> str:
        return self.user.visual_name or self.user.short_name

    @property
    def user_data(self):
        return None

    # endregion

    @property
    @abstractmethod
    def chat_id(self) -> Union[int, str]:
        pass

    @property
    @abstractmethod
    def message_id(self) -> Union[int, str]:
        pass

    @property
    def is_private(self) -> bool:
        return self.user_id == self.chat_id

    def get(self, key, default_value):
        return self._data.get(key, default_value)

    @property
    @abstractmethod
    def secret_code(self) -> str:
        pass

    def validate(self):
        assert self.user_id is not None


class InboundProcessor(ABC):
    error_message_default = 'Inbound processing error'

    @staticmethod
    @abstractmethod
    def process(context: BotContext, inbound: InboundMessage):
        """context.replies.add(BotMessage(inbound.text))"""
        pass


class DefaultInboundProcessor(InboundProcessor):
    """Empty processor"""

    @staticmethod
    def process(context: BotContext, inbound: InboundMessage):
        pass


class EchoInboundProcessor(InboundProcessor):
    """Echo processor"""

    @staticmethod
    def process(context: BotContext, inbound: InboundMessage):
        context.replies.add(BotMessage(inbound.text))


class EchoReplyToInboundProcessor(InboundProcessor):
    """Echo processor"""

    @staticmethod
    def process(context: BotContext, inbound: InboundMessage):
        context.replies.add(BotMessage(inbound.text, is_reply_to_message=True))


class EchoNameInboundProcessor(InboundProcessor):
    """Echo processor"""

    @staticmethod
    def process(context: BotContext, inbound: InboundMessage):
        # text = f'{context.bot.get_user_visual_name(inbound)} is your name.'
        text = f'{inbound.user.visual_name} is your name.'
        context.replies.add(BotMessage(text))


class InboundMessageFactory:
    __slots__ = ('_default_inbound_message_class',)

    def __init__(self, default_inbound_message_class: Type[InboundMessage]):
        self._default_inbound_message_class = default_inbound_message_class

    def create(self, data) -> InboundMessage:
        inbound_message_class = self.get_inbound_message_class(data)
        return inbound_message_class(data)

    def get_inbound_message_class(self, data) -> Type[InboundMessage]:
        return self._default_inbound_message_class



from abc import ABC, abstractmethod
from functools import cached_property
from typing import Union, Callable


class User(ABC):
    __slots__ = ('_uid',)

    def __init__(self, uid: Union[int, str], user_data_getter: Callable = None):
        self._uid = uid
        self._visual_name = None
        # self._user_data = None
        self._user_data_getter = user_data_getter

    def set_user_data_getter(self, getter_function: Callable):
        self._user_data_getter = getter_function

    @property
    def uid(self) -> Union[int, str]:
        return self._uid

    @cached_property
    def _data(self):
        return self._user_data_getter()

    @property
    @abstractmethod
    def data(self):
        """Must return self._data"""
        return self._data

    @cached_property
    def visual_name(self) -> str: return self.get_visual_name()

    @cached_property
    def short_name(self) -> str: return self.get_short_name()

    @abstractmethod
    def get_visual_name(self) -> str: pass

    @abstractmethod
    def get_short_name(self) -> str: pass

    @property
    @abstractmethod
    def id_url(self) -> str:
        pass



from typing import Union


class UBotHTTPResponse:
    def __init__(self,
                 content: Union[bytes, str],
                 code: int = 200,
                 **kwargs):
        self._content = content
        self._code = code
        self._kwargs = kwargs

    @property
    def content(self) -> Union[bytes, str]:
        return self._content

    @property
    def code(self) -> int:
        return self._code

    @property
    def params(self) -> dict:
        return self._kwargs


DEFAULT_SUCCESS_HTTP_RESPONSE = UBotHTTPResponse(b'ok')



class BotInfo:
    __slots__ = ('_name', '_data',)

    def __init__(self, name: str = '', **kwargs):
        self._name = name
        self._data = kwargs

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value



from typing import Type

from ubotadapter import InboundMessage, InboundMessageFactory
from .user import TelegramUser


class TGInboundMessage(InboundMessage):
    _UserClass = TelegramUser

    @property
    def message(self) -> dict:
        return self['message']

    @property
    def user_data(self) -> dict:
        return self._from

    @property
    def _from(self) -> dict:
        return self.message['from']

    @property
    def text_only(self) -> str:
        return self.message.get('text', '')

    @property
    def sticker(self) -> dict:
        return self.message.get('sticker', {})

    @property
    def text(self) -> str:
        return self.message.get('text', self.sticker.get('emoji', ''))

    @property
    def user_id(self) -> int:
        return self._from['id']

    @property
    def user_id_url(self) -> str:
        return f'@{self._from["username"]}'

    @property
    def chat_id(self) -> str:
        return self.message['chat']['id']

    @property
    def message_id(self) -> str:
        return self.message['message_id']

    @property
    def secret_code(self) -> None:
        return None

    @property
    def is_private(self) -> bool:
        return self.message('type', None) == 'private'


class TGInboundCallbackQuery(TGInboundMessage):
    @property
    def callback_query(self):
        return self['callback_query']

    @property
    def message(self):
        return self.callback_query['message']

    @property
    def _from(self):
        return self.callback_query['from']

    @property
    def text(self) -> str:
        data = self.callback_query['data']
        if isinstance(data, str):
            return data
        return ''


class TGInboundMessageFactory(InboundMessageFactory):
    def get_inbound_message_class(self, data) -> Type[InboundMessage]:
        if 'callback_query' in data:
            return TGInboundCallbackQuery
        return self._default_inbound_message_class



from typing import Generator, Any, Type, Dict, Union

from ubotadapter import UBotAdapter, InboundProcessor, ChannelType


class BotChannelCollection:
    def __init__(self):
        self._bots = {}

    def __iter__(self) -> Generator[UBotAdapter, Any, None]:
        return (bot for bot in self._bots.values())

    @property
    def bots(self) -> Dict[str, UBotAdapter]:
        return self._bots

    def __getitem__(self, bot_id_string: str) -> UBotAdapter:
        return self._bots[str(bot_id_string)]

    def __setitem__(self, bot_id_string: str, bot):
        bot_id_string = str(bot_id_string)
        if bot_id_string not in self._bots:
            self._bots[bot_id_string] = bot
        else:
            raise KeyError(f'Bot with id "{bot_id_string}" already exists in bot channel.')


class BotChannels:
    def __init__(self):
        self._channels = {}

    def __iter__(self) -> Generator[BotChannelCollection, Any, None]:
        return (collection for collection in self._channels.values())

    @property
    def channels(self) -> Dict[ChannelType, BotChannelCollection]:
        return self._channels

    def __getitem__(self, channel_type: ChannelType) -> BotChannelCollection:
        if channel_type not in self._channels:
            self._channels[channel_type] = BotChannelCollection()
        return self._channels[channel_type]

    def get_by_channel_name(self, channel_name: str) -> BotChannelCollection:
        try:
            channel = ChannelType[channel_name]
        except KeyError:
            raise KeyError(f'No ChannelType with name "{channel_name}" exists')
        else:
            return self[channel]


class BotsManager:
    def __init__(self):
        self._bots = {}
        self._channels = BotChannels()

    def __iter__(self) -> Generator[UBotAdapter, Any, None]:
        return (b for b in self.bots)

    def __getitem__(self, channel_type: Union[ChannelType, str]) -> BotChannelCollection:
        if isinstance(channel_type, ChannelType):
            return self.channels[channel_type]
        else:
            return self.channels.get_by_channel_name(channel_type)

    def add(self, bot: UBotAdapter):
        self._channels[bot.channel][bot.config.bot_id_string] = bot

    @property
    def bots(self) -> Generator[UBotAdapter, Any, None]:
        for channel in self.channels:
            for bot in channel:
                yield bot

    @property
    def channels(self) -> BotChannels:
        return self._channels

    def get_by_channel_name_and_id(self, channel_name: str, bot_id_string: str):
        channel = self.channels.get_by_channel_name(channel_name)
        return channel[str(bot_id_string)]

    def set_inbound_processor(self, inbound_processor: Type[InboundProcessor]):
        for bot in self.bots:
            bot.set_inbound_processor(inbound_processor)



import unittest

from ubotadapter.test.test_tools import *


class TestWebhooks(unittest.TestCase):
    def test_webhooks(self):
        for bot_test in BOTS_LIBRARY:
            bot = bot_test.bot
            with self.subTest(bot=bot.channel.name):
                # delete
                is_ok = bot.disable_webhook()
                self.assertTrue(is_ok, msg='Can\'t disable webhook')
                # check
                is_ok, url = bot.check_webhook()
                self.assertFalse(is_ok, msg='Webhook should be disabled')
                # update
                is_ok, url = bot.update_webhook()
                self.assertTrue(is_ok, msg=(is_ok, url))
                self.assertEqual(url, bot.config.webhook_url, msg=(is_ok, url))
                # delete
                is_ok = bot.disable_webhook()
                self.assertTrue(is_ok, msg='Can\'t disable webhook')


if __name__ == '__main__':
    unittest.main()



import json
import os
import random

# from ubotadapter.test.test_tools import *

# Test data JSON fields #
# %bot%:
#   bot_message_receiver:
# bot_default_message:
from abc import ABC, abstractmethod

_CUR_DIR = os.path.dirname(os.path.realpath(__file__))
DATA_DIR = os.path.join(_CUR_DIR, 'data')

with open(os.path.join(DATA_DIR, 'testdata.json'), 'r', encoding='utf-8') as f:
    _TEST_DATA = json.load(f)


def get_random_default_message():
    return f'{_TEST_DATA["bot_default_message"]} [{random.randint(100, 100000)}]'


class TestMessage(ABC):
    @property
    def default_text(self) -> str:
        return f'{_TEST_DATA["bot_default_message"]} [{random.randint(100, 100000)}]'

    def get_message(self, user_id: int = None, text: str = None, chat_id: int = None, bot_id: int = None, **kwargs):
        user_id = user_id or self.default_user_id
        text = text or self.default_text
        chat_id = chat_id or self.default_chat_id or user_id
        return self._get_message(user_id=user_id, text=text, chat_id=chat_id, **kwargs)

    @abstractmethod
    def _get_message(self, user_id: int = None, text: str = None, chat_id: int = None, bot_id: int = None, **kwargs):
        """Channel dependent message"""
        pass

    @property
    @abstractmethod
    def default_user_id(self) -> int:
        pass

    @property
    @abstractmethod
    def default_chat_id(self) -> int:
        pass


class VKTestMessage(TestMessage):
    @property
    def default_user_id(self) -> int:
        return _TEST_DATA['vk']['bot_message_receiver']

    @property
    def default_chat_id(self) -> int:
        return _TEST_DATA['vk']['default_chat_id']

    def _get_message(self, user_id: int = None, text: str = None, chat_id: int = None, bot_id: int = None, **kwargs):
        message_id = kwargs.get('message_id', 1)
        return {
            'type': 'message_new',
            'object': {
                'message': {
                    'date': 1624210000,
                    'from_id': user_id,
                    'id': message_id,
                    'out': 0,
                    'peer_id': chat_id,
                    'text': text,
                    'conversation_message_id': 1,
                    'fwd_messages': [],
                    'important': False,
                    'random_id': 0,
                    'attachments': [],
                    'is_hidden': False
                },
                'client_info': {
                    'button_actions': ['text', 'vkpay', 'open_app', 'location', 'open_link', 'callback',
                                       'intent_subscribe',
                                       'intent_unsubscribe'],
                    'keyboard': True,
                    'inline_keyboard': True,
                    'carousel': True,
                    'lang_id': 0
                }
            },
            'group_id': bot_id,
            'event_id': 'b6ca7d45d8055838911ed53a23f46b8098fe526f'
        }


class TGTestMessage(TestMessage):
    @property
    def default_user_id(self) -> int:
        return _TEST_DATA['telegram']['bot_message_receiver']

    @property
    def default_chat_id(self) -> int:
        return _TEST_DATA['telegram']['default_chat_id']

    def _get_message(self, user_id: int = None, text: str = None, chat_id: int = None, bot_id: int = None, **kwargs):
        return {
            'update_id': 350516797,
            'message': {
                'message_id': 1,
                'from': {
                    'id': user_id,
                    'is_bot': False,
                    'first_name': 'Igor',
                    'last_name': 'Shephard',
                    'username': 'smyek',
                    'language_code': 'en'
                },
                'chat': {
                    'id': chat_id,
                    'first_name': 'Igor',
                    'last_name': 'Shephard',
                    'username': 'smyek',
                    'type': 'private'
                },
                'date': 1624210000,
                'text': text
            }
        }


class TGCallbackTestMessage(TestMessage):
    @property
    def default_text(self) -> str:
        return f'callback_test [{random.randint(100, 100000)}]'

    @property
    def default_chat_id(self) -> int:
        return _TEST_DATA['telegram']['default_chat_id']

    @property
    def default_user_id(self) -> int:
        return _TEST_DATA['telegram']['bot_message_receiver']

    def _get_message(self, user_id: int = None, text: str = None, chat_id: int = None, bot_id: int = None, **kwargs):
        text = text or 'callback_test'
        return {
            'update_id': 350517096,
            'callback_query': {
                'id': '161535137947598551',
                'from': {
                    'id': user_id,
                    'is_bot': False,
                    'first_name': 'Igor',
                    'last_name': 'Shephard',
                    'username': 'smyek',
                    'language_code': 'en'
                },
                'message': {
                    'message_id': 581,
                    'from': {
                        'id': 1090207860,
                        'is_bot': True,
                        'first_name': 'STestBot',
                        'username': 'SmyekTestBot'
                    },
                    'chat': {
                        'id': chat_id,
                        'first_name': 'Igor',
                        'last_name': 'Shephard',
                        'username': 'smyek',
                        'type': 'private'
                    },
                    'date': 1627246822,
                    'text': 'Выбери',
                    'reply_markup': {
                        'inline_keyboard': [[{
                            'text': 'callback_test_button',
                            'callback_data': 'callback_test'
                        }
                        ]]
                    }
                },
                'chat_instance': '-4910036551047067769',
                'data': text
            }
        }


class DTFTestMessage(TestMessage):
    @property
    def default_user_id(self) -> int:
        return _TEST_DATA['dtf']['bot_message_receiver']

    @property
    def default_chat_id(self) -> int:
        return _TEST_DATA['dtf']['default_chat_id']

    def _get_message(self, user_id: int = None, text: str = None, chat_id: int = None, bot_id: int = None, **kwargs):
        text = text or get_random_default_message()
        test_entry_id = chat_id or 777326
        test_subsite_id = 130721
        reply_to = kwargs.get('reply_to', None)
        test_comment_id = kwargs.get('test_comment_id', 11825814)
        message = {
            "id": test_comment_id,
            "url": f'https://dtf.ru/{test_entry_id}?comment={test_comment_id}',
            "text": text,
            "media": [

            ],
            "creator": {
                "id": user_id,
                "avatar": "https://leonardo.osnova.io/d49d71ab-f78a-db1d-b4c4-0c72b8fcda0e/",
                "name": "Igor Shephard",
                "url": f'https://dtf.ru/u/{user_id}'
            },
            "content": {
                "id": test_entry_id,
                "title": "test",
                "url": f'https://dtf.ru/apitest/{test_entry_id}',
                "owner": {
                    "id": test_subsite_id,
                    "name": "Полигон",
                    "avatar": "",
                    "url": ""
                }
            },
            "reply_to": None
        }
        if reply_to:
            message['reply_to'] = {
                "id": reply_to,
                "url": "https://",
                "text": "Это родительский комментарий",
                "media": [

                ],
                "creator": {
                    "id": 1,
                    "avatar": "https://",
                    "name": "TestUser",
                    "url": ""
                }
            }
        return message



from ubotadapter.bots import Telegram
from ubotadapter.inbound import EchoInboundProcessor

def test_msg():
    tgbot = Telegram('data/telegram.json')
    tgbot.set_inbound_processor(EchoInboundProcessor)
    response = tgbot.handle({
        'update_id': 350516797,
        'message': {
            'message_id': 9,
            'from': {
                'id': 37610330,
                'is_bot': False,
                'first_name': 'Igor',
                'last_name': 'Shephard',
                'username': 'smyek',
                'language_code': 'en'
            },
            'chat': {
                'id': 37610330,
                'first_name': 'Igor',
                'last_name': 'Shephard',
                'username': 'smyek',
                'type': 'private'
            },
            'date': 1622413728,
            # 'text': 'reply_to_message'
            'text': 'test_text'
        }
    })
    return response


if __name__ == '__main__':
    import logging
    logging.getLogger('ubotadapter').addHandler(logging.StreamHandler())
    test_msg()



from typing import List

from ubotadapter import BotContext, InboundMessage, BotMessage
from ubotadapter.botmsg import Keyboard
from ubotadapter.inbound import InboundProcessor

MARKER = 'utest_cmnd_in'


def wrap_buttons(buttons: List[List[str]]):
    for i in range(len(buttons)):
        for j in range(len(buttons[i])):
            buttons[i][j] = f'{MARKER} {buttons[i][j]}'
    return buttons


class TestingInboundProcessor(InboundProcessor):
    @staticmethod
    def process(context: BotContext, inbound: InboundMessage):
        text = inbound.text.lower()
        tokens = text.split(' ')
        set_tokens = set(tokens)
        if MARKER not in set_tokens:
            return

        reply_text = ''
        keyboard = None
        data = {}

        if 'echo' in set_tokens:
            reply_text = inbound.text
        elif 'say my name' in text:
            reply_text = f'OK – {inbound.user.visual_name}'
        elif 'reply to me' in text:
            reply_text = f'Replying.'
            data['is_reply_to_message'] = True
        elif 'buttons' in set_tokens:
            in_line = 'in_line' in set_tokens or 'inline' in set_tokens
            reply_text = 'Выбери'
            buttons = [['echo', 'say my name'], ['buttons', 'reply to me']]
            buttons = wrap_buttons(buttons)
            keyboard = Keyboard(buttons=buttons, one_time=False, in_line=in_line)

        if reply_text:
            data['text'] = reply_text
            if keyboard:
                data['keyboard'] = keyboard
            message = BotMessage(**data)
            context.replies.add(message)





from typing import List, Generator, Any, Callable


class Keyboard:
    def __init__(self,
                 buttons: List[List[str]],
                 n_cols: int = 1,
                 in_line: bool = False,
                 one_time: bool = False):
        self.buttons = buttons
        self.n_cols = n_cols
        self.in_line = in_line
        self.one_time = one_time

    def rows(self) -> Generator[List[str], Any, None]:
        return (row for row in self.buttons)

    def __iter__(self) -> Generator[List[str], Any, None]:
        yield self.rows

    def serialize_buttons(self, serializer: Callable):
        return ([serializer(button) for button in row] for row in self.buttons)


class BotMessage:
    __slots__ = ('is_reply_to_message', 'reply_to_message', 'data', '_text', '_keyboard')

    def __init__(self, text: str,
                 is_reply_to_message: bool = False,
                 keyboard: Keyboard = None,
                 reply_to_message=None,
                 **kwargs):
        self._text = text
        self._keyboard = keyboard
        self.is_reply_to_message = is_reply_to_message
        self.reply_to_message = reply_to_message
        self.data = kwargs

    @property
    def content(self):
        return self.text

    @property
    def text(self) -> str:
        return self._text

    @property
    def keyboard(self) -> Keyboard:
        return self._keyboard

    @property
    def has_content(self) -> bool:
        return bool(self._text or self._keyboard)


class BotReplies:
    __slots__ = ('_messages',)

    def __init__(self):
        self._messages: List[BotMessage] = []

    def __iter__(self) -> Generator[BotMessage, Any, None]:
        return (m for m in self._messages)

    def __getitem__(self, idx) -> BotMessage:
        return self._messages[idx]

    @property
    def count(self) -> int:
        return len(self._messages)

    def add(self, message: BotMessage):
        self._messages.append(message)

    def add_text(self, text: str):
        self._messages.append(BotMessage(text))

    def clear(self):
        self._messages.clear()



from ubotadapter import InboundMessage

from .user import VKUser


class VKInboundMessage(InboundMessage):
    _UserClass = VKUser

    @property
    def object(self):
        return self['object']

    @property
    def message(self):
        return self.object['message']

    @property
    def group_id(self) -> int:
        return self.get('group_id', None)

    @property
    def text(self) -> str:
        return self.message['text']

    @property
    def user_id(self) -> str:
        return self.message['from_id']

    @property
    def user_id_url(self) -> str:
        return f'@id{self.user_id}'

    @property
    def chat_id(self) -> str:
        return self.message['peer_id']

    @property
    def message_id(self) -> int:
        return self.message['id']

    @property
    def vk_conversation_message_id(self) -> int:
        return self.message['conversation_message_id']

    @property
    def secret_code(self) -> str:
        return self.get('secret', None)



import json

from vkale import VKaleAPI

from ubotadapter import UBotAdapter, BotInfo, ChannelType
from ubotadapter.botdebug import *
from ubotadapter.http import *
from ubotadapter.logger import logger
from .inbound import VKInboundMessage


class VK(UBotAdapter):
    channel = ChannelType.VK
    _default_server_name = 'main'
    _InboundMessageClass = VKInboundMessage
    _info = BotInfo(channel=ChannelType.VK)

    def _init_api(self):
        self._api = VKaleAPI(self.config.token)

        if self.config.get('server_name'):
            self._default_server_name = self.config['server_name']

    @property
    def api(self) -> VKaleAPI:
        return self._api

    def set_webhook(self):
        server = self.api.groups.get_callback_server_by_name(group_id=self.config.bot_id,
                                                             server_name=self._default_server_name)
        if not server:
            server_id = self.api.groups.add_callback_server(group_id=self.config.bot_id,
                                                            url=self.config.webhook_url,
                                                            title=self._default_server_name,
                                                            secret_key=self.config.webhook_secret)
            server = self.api.groups.get_callback_servers(group_id=self.config.bot_id, servers_id=server_id)[0]
        response = self.api.groups('editCallbackServer',
                                   group_id=self.config.bot_id,
                                   server_id=server.id,
                                   url=self.config.webhook_url,
                                   title=self._default_server_name,
                                   secret_key=self.config.webhook_secret)
        response = self.api.groups.set_callback_settings(group_id=self.config.bot_id,
                                                         server_id=server.id,
                                                         message_new=1)

        if not response:
            raise Exception(f'Cant add server {response}')

        return response

    def disable_webhook(self) -> bool:
        server = self.api.groups.get_callback_server_by_name(group_id=self.config.bot_id,
                                                             server_name=self._default_server_name)
        if server:
            response = self.api.groups.set_callback_settings(group_id=self.config.bot_id,
                                                             server_id=server.id,
                                                             message_new=0)
            return response == 1
        return True

    def check_webhook(self) -> (bool, str):
        server = self.api.groups.get_callback_server_by_name(group_id=self.config.bot_id,
                                                             server_name=self._default_server_name)
        if not server:
            return False, ''
        settings = self.api.groups.get_callback_settings(group_id=self.config.bot_id,
                                                         server_id=server.id)
        status = server.url == self.config.webhook_url and settings.message_new == 1  # server.status == 'ok'
        return status, server.url

    def reset_webhook(self) -> (bool, str):
        self.disable_webhook()
        return self.update_webhook()

    def update_webhook(self) -> (bool, str):
        is_ok, url = self.check_webhook()
        if not is_ok or url != self.config.webhook_url:
            self.set_webhook()
            return self.check_webhook()
        return is_ok, url

    def reply(self, context, message):
        def serializer_button(text):
            return {
                "action": {
                    "type": "text",
                    # "payload": "{\"button\": \"1\"}",
                    "label": text
                },
                "color": "secondary"
            }

        params = dict(peer_id=context.inbound.chat_id,
                      message=message.text,
                      random_id=0)

        if message.is_reply_to_message:
            message_id = message.reply_to_message if message.reply_to_message else self.context.inbound.message_id
            if self.context.inbound.is_private:
                params['reply_to'] = message_id

        if message.keyboard:
            buttons = list(message.keyboard.serialize_buttons(serializer_button))
            keyboard = {
                'buttons': buttons,
                'inline': message.keyboard.in_line,
                'one_time': message.keyboard.one_time and not message.keyboard.in_line
            }
            params['keyboard'] = json.dumps(keyboard)

        response = self.api.method(method_name='messages.send', **params)
        # Success
        if response and 'response' in response:
            if isinstance(response['response'], int):
                return True, BotReplyStatus.ok
        # Errors
        logger.error(f'VK error response: {response}')
        if 'error' in response:
            code = response['error']['error_code']
            if code == 901:
                return False, BotReplyStatus.no_permission
            raise ValueError(response)
        return False, BotReplyStatus.unknown_error

    def get_confirmation_code(self) -> str:
        code = self.api.groups.get_callback_confirmation_code(group_id=self.config.bot_id)
        return code

    def get_user_data(self, inbound: VKInboundMessage):
        return self.api.users.get(user_ids=inbound.user_id)[0]

    def get_prehandle_webhook_response(self, data: dict) -> Union[UBotHTTPResponse, None]:
        if data.get('type', None) == 'confirmation':
            logger.info('vk confirmation')
            return UBotHTTPResponse(self.get_confirmation_code(), status=200, is_forced_response=True)
        return None


if __name__ == '__main__':
    vkubot = VK('config/vk.json')
    print(vkubot.check_webhook())



from deteefapi import DeteefAPI
from deteefapi.webhook import CommentCreator

from ubotadapter.user import User


class DTFUser(User):
    @property
    def data(self) -> CommentCreator:
        return super()._data

    def get_visual_name(self) -> str:
        return self.data.name

    def get_short_name(self) -> str:
        return self.data.name

    @property
    def id_url(self) -> str:
        return DeteefAPI.get_user_url(self.data.id, self.visual_name)



from .botadapter import UBotAdapter
from .botcontext import BotContext
from .botdebug import *
from .botinfo import BotInfo
from .botmsg import BotMessage
from .enums import ChannelType
from .inbound import InboundProcessor, InboundMessage, InboundMessageFactory
from .manager import BotsManager
from .msgfilter import *



import setuptools

setuptools.setup(name='ubotadapter',
                 version='0.3',
                 description='Bot Adapter class',
                 url='',
                 author='Smyek',
                 author_email='smyek.job@gmail.com',
                 license='MIT',
                 packages=setuptools.find_packages(),
                 install_requires=['backoff', 'vkale>=0.4.2', 'python-telegram-bot', 'deteefapi>=0.4.1'],
                 zip_safe=False)



import json
import traceback
from abc import ABC, abstractmethod
from functools import partial
from typing import Union, Type

import backoff as backoff

from .botconfig import BotConfig
from .botcontext import BotContext
from .botdebug import *
from .botinfo import BotInfo
from .botmsg import BotMessage
from .enums import ChannelType
from .inbound import InboundProcessor, InboundMessage, DefaultInboundProcessor, InboundMessageFactory
from .logger import logger
from .msgfilter import MessageFilterManager, NotBotMessage
from .http import *


class UBotAdapter(ABC):
    channel: ChannelType = None
    _info = None
    _BotContextClass = BotContext
    _BotMessageClass = BotMessage
    _inbound_message_factory = InboundMessageFactory
    _InboundMessageClass = InboundMessage
    _InboundProcessorClass = DefaultInboundProcessor
    _success_webhook_response = DEFAULT_SUCCESS_HTTP_RESPONSE

    __slots__ = ('_api', '_config', '_context', '_inbound_processor', '_message_filters_manager')

    def __init__(self, config: Union[str, dict, BotConfig]):
        """config. str == path, dict -> config"""
        if isinstance(config, str):
            config = BotConfig(self._load_config(config))
        elif isinstance(config, dict):
            config = BotConfig(config)
        self._config = config
        # info
        self.info.name = self.config.name or f'{self.channel.name}_{self.config.bot_id}'
        # API
        self._api = None
        self._inbound_processor = self._InboundProcessorClass
        self._init_api()
        self._message_filters_manager = MessageFilterManager([NotBotMessage])
        self._context = self._BotContextClass(self)
        logger.debug(f'Bot "{self.channel.name}" with name "{self.info.name}" initialized')

    @property
    @abstractmethod
    def api(self):
        return self._api

    @property
    def context(self) -> _BotContextClass:
        return self._context

    @property
    def inbound_message_factory(self):
        return self._inbound_message_factory(self._InboundMessageClass)

    def _init_api(self):
        return None

    @property
    def info(self) -> BotInfo:
        return self._info

    @property
    def message_filters_manager(self) -> MessageFilterManager:
        return self._message_filters_manager

    @staticmethod
    def _load_config(path) -> dict:
        """
        token: str
        webhook_url: str
        confirmation_code (optional): str
        :param str path:
        :return: dict config
        """
        with open(path, 'r') as f:
            return json.load(f)

    # region Webhooks
    @abstractmethod
    def check_webhook(self) -> (bool, str):
        """Check if webhook is set"""
        pass

    @property
    def is_webhook_set(self) -> bool:
        """Check if webhook is online"""
        is_ok, url = self.check_webhook()
        return is_ok

    @abstractmethod
    def set_webhook(self) -> bool:
        pass

    @abstractmethod
    def disable_webhook(self) -> bool:
        pass

    @abstractmethod
    def reset_webhook(self) -> (bool, str):
        pass

    @abstractmethod
    def update_webhook(self) -> (bool, str):
        """Check if webhook is relevant, is set and active, if not - set"""
        pass

    # endregion

    @abstractmethod
    def reply(self, context: BotContext, message: BotMessage) -> (bool, BotReplyStatus):
        """Returns (True, BotReplyStatus.ok) if success, (False, BotReplyStatus.<error>) if not"""
        pass

    @property
    def config(self) -> BotConfig:
        return self._config

    @property
    def inbound_processor(self) -> Type[InboundProcessor]:
        return self._inbound_processor

    def handle(self, data) -> BotHandleResult:
        self.context.clear()  # TODO: is it necessary?

        logger.debug('serialize inbound')
        inbound = self._init_inbound(data)

        # process inbound message
        logger.debug('processing')
        if inbound:
            if inbound.secret_code and self.config.webhook_secret != inbound.secret_code:
                data = {'webhook': self.config.webhook_secret, 'inbound': inbound.secret_code}
                self.context.handle_result.process_result = BotProcessResult(BotProcessStatus.wrong_secret, data=data)
            else:
                self.context.handle_result.process_result = self._process(inbound)

        # reply
        logger.debug('replying')
        if not self.context.replies.count:
            logger.info('No messages in replies')
        for message in self.context.replies:
            if self.is_message_valid(message):
                self.context.handle_result.reply_result = self._reply(message)
            else:
                logger.info('Warning: no message')
                self.context.handle_result.reply_result = BotReplyResult(BotReplyStatus.ok)

        return self.context.handle_result

    @property
    def success_webhook_response(self) -> UBotHTTPResponse:
        return self._success_webhook_response

    def get_prehandle_webhook_response(self, data) -> Union[UBotHTTPResponse, None]:
        """returned response should be sent before main processing"""
        pass

    def set_inbound_processor(self, processor: Type[InboundProcessor]):
        self._inbound_processor = processor

    def _init_inbound(self, data) -> InboundMessage:
        inbound = None
        try:
            inbound = self.inbound_message_factory.create(data)
            inbound.user.set_user_data_getter(partial(self.get_user_data, inbound))
            self.context.inbound = inbound  # TODO: why set inbound?
        except Exception as e:
            self.context.handle_result.process_result = BotProcessResult(BotProcessStatus.inbound_error,
                                                                         exception=traceback.format_exc())
        return inbound

    def _process(self, inbound: InboundMessage) -> BotProcessResult:
        # filtering
        not_passed_filter = self.message_filters_manager.check(self.context, inbound)
        if not_passed_filter is not None:
            return BotProcessResult(BotProcessStatus.filter_not_passed, data=not_passed_filter.__name__)
        # main process
        try:
            result = self.inbound_processor.process(self.context, inbound)
            return BotProcessResult(BotProcessStatus.ok, data=result)
        except Exception as e:
            logger.error(f'Failed on _process with exception: {repr(e)}')
            if self.config.is_send_process_error_message:
                self.context.replies.add(BotMessage(self.inbound_processor.error_message_default))
            return BotProcessResult(BotProcessStatus.unknown_error, exception=traceback.format_exc())

    @backoff.on_exception(backoff.expo,
                          BotReplyException,
                          max_tries=10,
                          max_time=60)
    def _reply(self, message) -> (bool, BotReplyStatus):
        try:
            is_ok, reply_status = self.reply(self.context, message)
            if not is_ok:
                logger.error(f'No reply: {reply_status}')
            return BotReplyResult(reply_status, exception=reply_status)
        except Exception as e:
            error_msg = f'Failed on _process with exception: {repr(e)}'
            logger.error(error_msg, exc_info=1)
            return BotReplyResult(BotReplyStatus.unknown_exception, exception=traceback.format_exc())

    def is_message_valid(self, message: BotMessage) -> bool:
        return isinstance(message, BotMessage) and message.has_content

    @abstractmethod
    def get_user_data(self, inbound: InboundMessage):
        pass



import itertools

from deteefapi import DeteefAPI
from deteefapi.webhook import CommentCreator

from ubotadapter import UBotAdapter, BotInfo, ChannelType, BotMessage, BotReplyStatus, BotContext
from ubotadapter.logger import logger
from .inbound import DTFInboundMessage


class DTF(UBotAdapter):
    channel = ChannelType.DTF
    _info = BotInfo(name=ChannelType.DTF.name)
    _InboundMessageClass = DTFInboundMessage

    def _init_api(self):
        self._api = DeteefAPI(self.config.token)

    @property
    def api(self) -> DeteefAPI:
        return self._api

    def check_webhook(self) -> (bool, str):
        webhook_data = self.api.get_webhook()
        is_set = webhook_data.is_webhook(self.config.webhook_url, event_type='new_comment')
        webhook_url = webhook_data.get_url()
        if is_set and webhook_url == self.config.webhook_url:
            return True, webhook_url
        return False, webhook_url

    def set_webhook(self) -> bool:
        response = self.api.set_webhook(self.config.webhook_url, event='new_comment')
        result = response.json()['result']
        return response.status_code == 200 and result['url'] == self.config.webhook_url

    def disable_webhook(self) -> bool:
        response = self.api.delete_webhook(webhook_url=self.config.webhook_url, event='new_comment')
        result = response.json()['result']
        return response.status_code == 200 and result['success']

    def reset_webhook(self) -> (bool, str):
        webhook_data = self.api.get_webhook()
        webhook_url = webhook_data.get_url()
        if webhook_url:
            self.api.delete_webhook(webhook_url=webhook_url, event='new_comment')
        is_ok, url = self.set_webhook(), self.config.webhook_url
        return is_ok, url

    def update_webhook(self) -> (bool, str):
        is_ok, url = self.check_webhook()
        if not is_ok:
            self.reset_webhook()
        return self.check_webhook()

    def reply(self, context: BotContext, message: BotMessage) -> (bool, BotReplyStatus):
        text = message.text
        post_id = context.inbound.chat_id
        reply_to = context.inbound.message_id
        if message.keyboard:
            buttons_chain = itertools.chain.from_iterable(message.keyboard.buttons)
            buttons_text_list = [f'{i + 1}. {btn}' for i, btn in enumerate(buttons_chain)]
            buttons_text = '\n'.join(buttons_text_list)
            if not text:
                text = buttons_text
            else:
                text += '\n' + buttons_text
        response = self.api.comment_send(post_id=post_id, text=text, reply_to=reply_to)
        if response.status_code == 200:
            return True, BotReplyStatus.ok
        logger.error(f'DTF error response: {response.json()}')
        return False, BotReplyStatus.unknown_error

    def get_user_data(self, inbound: DTFInboundMessage) -> CommentCreator:
        return inbound.user_data



from functools import cached_property

from deteefapi import WebhookComment, DeteefAPI
from deteefapi.webhook import CommentCreator

from ubotadapter import InboundMessage
from .user import DTFUser


class DTFInboundMessage(InboundMessage):
    _UserClass = DTFUser

    @property
    def user_id(self) -> int:
        return self._comment.creator.id

    @property
    def user_id_url(self) -> str:
        return DeteefAPI.get_user_url(user_id=self.user_id, name=self._comment.creator.name)

    @property
    def chat_id(self) -> int:
        return self._comment.content.id

    @property
    def message_id(self) -> int:
        return self._comment.id

    @property
    def secret_code(self) -> str:
        # TODO: implement
        return ''

    @property
    def text(self) -> str:
        return self._comment.text

    @cached_property
    def _comment(self) -> WebhookComment:
        return WebhookComment(self._data)

    @property
    def user_data(self) -> CommentCreator:
        return self._comment.creator



from enum import Enum


class ChannelType(Enum):
    Telegram = 0
    VK = 1
    Discord = 2  # not yet supported
    WhatsApp = 10  # not yet supported
    WeChat = 11  # not yet supported
    Skype = 20  # not yet supported
    DTF = 100
